/* 
 *
 * COPYRIGHT NVIDIA CORPORATION 2002. ALL RIGHTS RESERVED.  
 * BY ACCESSING OR USING THIS SOFTWARE, YOU AGREE TO: 
 *
 *  1) ACKNOWLEDGE NVIDIA'S EXCLUSIVE OWNERSHIP OF ALL RIGHTS 
 *     IN AND TO THE SOFTWARE; 
 *
 *  2) NOT MAKE OR DISTRIBUTE COPIES OF THE SOFTWARE WITHOUT 
 *     INCLUDING THIS NOTICE AND AGREEMENT; 
 *
 *  3) ACKNOWLEDGE THAT TO THE MAXIMUM EXTENT PERMITTED BY 
 *     APPLICABLE LAW, THIS SOFTWARE IS PROVIDED *AS IS* AND 
 *     THAT NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, 
 *     EITHER EXPRESS OR IMPLIED, INCLUDING, BUT NOT LIMITED
 *     TO, IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS 
 *     FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS BE LIABLE FOR ANY 
 * SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES 
 * WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS 
 * OF BUSINESS PROFITS, BUSINESS INTERRUPTION, LOSS OF BUSINESS 
 * INFORMATION, OR ANY OTHER PECUNIARY LOSS), INCLUDING ATTORNEYS'
 * FEES, RELATING TO THE USE OF OR INABILITY TO USE THIS SOFTWARE, 
 * EVEN IF NVIDIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *
 */

/*
 *
 * NVidia Corporation
 *
 * Written by Ryan Geiss
 *
 * These helper functions let you reload shaders on the fly.
 * Note that whenever you reload a shader, you have to flush the bindcache
 * on all the shaderparams it uses.  Some of these params are directly
 * under the shader (GlobalParams, DefaultParams); some of them
 * are on the shapes (FragmentShaderParams, VertexShaderParams); 
 * and some of them are scattered throughout NVDemo (such as in the 
 * case of the NVMenuItem stuff).
 *
 */



#include "ReloadShaders.h"



class NVReloadAllShaderTrav : public NVTraversal {
public:
  // constructor
  NVReloadAllShaderTrav(int n) {};

  // for every NVShaderParam in the scene, flush its cache
  virtual bool Apply(NVNode *node){
    NVShaderParam* p = dynamic_cast<NVShaderParam*>(node);
    if (p)
      p->FlushCache();
    NVShader* s = dynamic_cast<NVShader*>(node);
    if (s)
      s->ReloadShader();

    // hacks for any non-conforming classes: [...that have NVNode-derived members that aren't wrapped up in fields!]
    NVDropDownMenu* m = dynamic_cast<NVDropDownMenu*>(node);
    if (m && m->m_clickable) 
      for (uint i=0; i<m->m_clickable->GetFragmentShaderParams().size(); i++)
        m->m_clickable->GetFragmentShaderParams()[i]->FlushCache();

    return false; // continue the traversal
  }
};



class NVReloadSomeShaderTrav : public NVTraversal {
public:
  // data members
  NVShader **m_shaders;
  int      m_nShaders;

  // constructor
  NVReloadSomeShaderTrav(NVShader **shaders, int numShaders):
    m_shaders(shaders),
    m_nShaders(numShaders)
  { };

  // for every shape in the scene, if its vertex or fragment shader are
  // being reloaded, flush the cache on all the shape's params.
  virtual bool Apply(NVNode *node){
    NVShape* s = dynamic_cast<NVShape*>(node);
    if (s) {
      NVRenderSettings* rs = dynamic_cast<NVRenderSettings*>(s->GetRenderSettings());
      if (rs) {
        bool bFlushShapeParams[2] = { false, false };
        NVShader *pVS = rs->GetVertexShader();
        NVShader *pFS = rs->GetFragmentShader();
        for (int i=0; i<m_nShaders; i++) {
          if (pVS == m_shaders[i])
            bFlushShapeParams[0] = true;
          if (pFS == m_shaders[i]) 
            bFlushShapeParams[1] = true;
        }
        if (bFlushShapeParams[0])
          for (int i=s->GetVertexShaderParams().size()-1; i>=0; i--)
            s->GetVertexShaderParams()[i]->FlushCache();
        if (bFlushShapeParams[1])
          for (int i=s->GetFragmentShaderParams().size()-1; i>=0; i--)
            s->GetFragmentShaderParams()[i]->FlushCache();
      }
    }
    return false; // continue the traversal
  }
};



void ReloadAllShaders(NVScene* scene)
{
  // 1. reload ALL shaders
  /*
  const std::vector<NVShader*> &shaders = scene->GetShaders();
  uint numShaders = shaders.size();
  for (uint i = 0; i < numShaders; i++)
    shaders[i]->ReloadShader();
  */

  // 2. also have to traverse ALL NVShaderParam nodes to reset their bindcaches.
  //    if this fails to catch them all and something goes wrong, it's probably
  //    because some class has an NVShaderParam that isn't part of the scene
  //    hierarchy (i.e. it's a normal class member, instead of an SNODE or MNODE).
  NVReloadAllShaderTrav trav(0);
  scene->Traverse(&trav);
}



void ReloadSomeShaders(NVScene *scene, std::string *shaders_to_reload, int N)
{
  // 1. reload just a few shaders, and manually flush the caches on all their params
  NVShader* shaders[256];
  assert(N < sizeof(shaders)/sizeof(NVShader*));
  for (int i=0; i<N; i++) {
    shaders[i] = dynamic_cast<NVShader*>(NVNAMESPACE->GetByName(shaders_to_reload[i]));
    if (!shaders[i]) {
      NOutputString("ReloadSomeShaders: WARNING: shader " + shaders_to_reload[i] + " not found\n");
      continue;
    }
    shaders[i]->ReloadShader();

    int N = shaders[i]->GetGlobalParams().size();
    for (int j=0; j<N; j++)
      shaders[i]->GetGlobalParams()[j]->FlushCache();
    N = shaders[i]->GetDefaultParams().size();
    for (int j=0; j<N; j++)
      shaders[i]->GetDefaultParams()[j]->FlushCache();
  }
  
  // 3. find all shapes that use any of these shaders & reset the bindcaches
  //    on those SHAPES' VS/FS params.
  NVReloadSomeShaderTrav trav(shaders, N);
  scene->Traverse(&trav);

}

void ReloadSomeShaders(NVScene *scene, std::vector<std::string> &shaders_to_reload)
{
  // 1. reload just a few shaders, and manually flush the caches on all their params
  NVShader* shaders[256];
  for (int i=0; i<shaders_to_reload.size(); i++) {
    shaders[i] = dynamic_cast<NVShader*>(NVNAMESPACE->GetByName(shaders_to_reload[i]));
    if (!shaders[i]) {
      NOutputString("ReloadSomeShaders: WARNING: shader " + shaders_to_reload[i] + " not found\n");
      continue;
    }
    shaders[i]->ReloadShader();

    int N = shaders[i]->GetGlobalParams().size();
    for (int j=0; j<N; j++)
      shaders[i]->GetGlobalParams()[j]->FlushCache();
    N = shaders[i]->GetDefaultParams().size();
    for (int j=0; j<N; j++)
      shaders[i]->GetDefaultParams()[j]->FlushCache();
  }
  
  // 3. find all shapes that use any of these shaders & reset the bindcaches
  //    on those SHAPES' VS/FS params.
  NVReloadSomeShaderTrav trav(shaders, shaders_to_reload.size());
  scene->Traverse(&trav);

}