/* 
 *
 * COPYRIGHT NVIDIA CORPORATION 2002. ALL RIGHTS RESERVED.  
 * BY ACCESSING OR USING THIS SOFTWARE, YOU AGREE TO: 
 *
 *  1) ACKNOWLEDGE NVIDIA'S EXCLUSIVE OWNERSHIP OF ALL RIGHTS 
 *     IN AND TO THE SOFTWARE; 
 *
 *  2) NOT MAKE OR DISTRIBUTE COPIES OF THE SOFTWARE WITHOUT 
 *     INCLUDING THIS NOTICE AND AGREEMENT; 
 *
 *  3) ACKNOWLEDGE THAT TO THE MAXIMUM EXTENT PERMITTED BY 
 *     APPLICABLE LAW, THIS SOFTWARE IS PROVIDED *AS IS* AND 
 *     THAT NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, 
 *     EITHER EXPRESS OR IMPLIED, INCLUDING, BUT NOT LIMITED
 *     TO, IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS 
 *     FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS BE LIABLE FOR ANY 
 * SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES 
 * WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS 
 * OF BUSINESS PROFITS, BUSINESS INTERRUPTION, LOSS OF BUSINESS 
 * INFORMATION, OR ANY OTHER PECUNIARY LOSS), INCLUDING ATTORNEYS'
 * FEES, RELATING TO THE USE OF OR INABILITY TO USE THIS SOFTWARE, 
 * EVEN IF NVIDIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *
 */

#include "support.h"
#include <windows.h>

float AdjustRateToFPS(float per_frame_decay_rate_at_fps1, float fps1, float actual_fps)
{
    // returns the equivalent per-frame decay rate at actual_fps
    
    // basically, do all your testing at fps1 and get a good decay rate;
    // then, in the real application, adjust that rate by the actual fps each time you use it.
    
    float per_second_decay_rate_at_fps1 = powf(per_frame_decay_rate_at_fps1, fps1);
    float per_frame_decay_rate_at_fps2 = powf(per_second_decay_rate_at_fps1, 1.0f/actual_fps);

    return per_frame_decay_rate_at_fps2;
}

NVec3f TransformProjective(const NVec3f &pt, const NMatf &m)
{
  const float* f = m[0];
  float x = pt.x*f[ 0] + pt.y*f[ 1] + pt.z*f[ 2] + f[ 3];
  float y = pt.x*f[ 4] + pt.y*f[ 5] + pt.z*f[ 6] + f[ 7];
  float z = pt.x*f[ 8] + pt.y*f[ 9] + pt.z*f[10] + f[11];
  float w = pt.x*f[12] + pt.y*f[13] + pt.z*f[14] + f[15];
  w = 1/w;
  NVec3f ret = NVec3f(x*w, y*w, z*w);
  return ret;
}

NPnt3f TransformProjective(const NPnt3f &pt, const NMatf &m)
{
  const float* f = m[0];
  float x = pt.x*f[ 0] + pt.y*f[ 1] + pt.z*f[ 2] + f[ 3];
  float y = pt.x*f[ 4] + pt.y*f[ 5] + pt.z*f[ 6] + f[ 7];
  float z = pt.x*f[ 8] + pt.y*f[ 9] + pt.z*f[10] + f[11];
  float w = pt.x*f[12] + pt.y*f[13] + pt.z*f[14] + f[15];
  w = 1/w;
  NPnt3f ret = NPnt3f(x*w, y*w, z*w);
  return ret;
}

NVec3f TransformNonProjective(const NVec3f &pt, const NMatf &m)
{
  const float* f = m[0];
  float x = pt.x*f[ 0] + pt.y*f[ 1] + pt.z*f[ 2] + f[ 3];
  float y = pt.x*f[ 4] + pt.y*f[ 5] + pt.z*f[ 6] + f[ 7];
  float z = pt.x*f[ 8] + pt.y*f[ 9] + pt.z*f[10] + f[11];
  NVec3f ret = NVec3f(x, y, z);
  return ret;
}

NPnt3f TransformNonProjective(const NPnt3f &pt, const NMatf &m)
{
  const float* f = m[0];
  float x = pt.x*f[ 0] + pt.y*f[ 1] + pt.z*f[ 2] + f[ 3];
  float y = pt.x*f[ 4] + pt.y*f[ 5] + pt.z*f[ 6] + f[ 7];
  float z = pt.x*f[ 8] + pt.y*f[ 9] + pt.z*f[10] + f[11];
  NPnt3f ret = NPnt3f(x, y, z);
  return ret;
}

NVec3f Transform3x3(const NVec3f &pt, const NMatf &m)  // no translate!
{
  const float* f = m[0];
  float x = pt.x*f[ 0] + pt.y*f[ 1] + pt.z*f[ 2];
  float y = pt.x*f[ 4] + pt.y*f[ 5] + pt.z*f[ 6];
  float z = pt.x*f[ 8] + pt.y*f[ 9] + pt.z*f[10];
  NVec3f ret = NVec3f(x, y, z);
  return ret;
}

void ScaleVerts(NVBaseMesh *baseMesh, const NVec3f &scale){
  // get the object space verts
  NVPnt3fArrayNode *coordNode = (NVPnt3fArrayNode*) baseMesh->FindChannel(NV_A2V_OBJ_COORD);
  assert(ISA(coordNode, NVPnt3fArrayNode*));
  
  std::vector<NPnt3f> coords = coordNode->GetValue();
  uint numNormals = coords.size();
  for (uint ni = 0; ni < numNormals; ni++) {
    coords[ni].x = coords[ni].x * scale.x;
    coords[ni].y = coords[ni].y * scale.y;
    coords[ni].z = coords[ni].z * scale.z;
  }

  coordNode->SetValue(coords);
}

NVec3f GetMatrixTranslation(const NMatf &m) 
{
  const float* f = m[0]; 
  return NVec3f(f[3], f[7], f[11]); 
}


//------------[ REGISTRY HELPER FUNCTIONS ]------------

// 's_bCurrentUserOnly': extern this and set it to TRUE if you want your demo
// to save settings on a per-user basis!  (vs. local machine)
bool s_bCurrentUserOnly = false;

int NVGetIntFromReg(const char* szDemoName, const char* szValueName, int defaultValue)
{
  char szKey[512];
  sprintf(szKey, "SOFTWARE\\NVIDIA Corporation\\NVIDIA Demos\\%s", szDemoName);

  LONG x;
  HKEY hk; 
  x = RegOpenKeyEx(s_bCurrentUserOnly ? HKEY_CURRENT_USER : HKEY_LOCAL_MACHINE, szKey, 
          0, KEY_READ, &hk);
  if (x != ERROR_SUCCESS)
    return defaultValue;

  DWORD type;
  int ret;
  DWORD ret_size = sizeof(ret);
  x = RegQueryValueEx(hk, szValueName, NULL, &type, (LPBYTE)(&ret), &ret_size);
  RegCloseKey(hk);
  if (x != ERROR_SUCCESS || type != REG_DWORD)
    return defaultValue;
  return ret;
}

bool NVWriteIntToReg(const char* szDemoName, const char* szValueName, int value)
{
  char szKey[512];
  sprintf(szKey, "SOFTWARE\\NVIDIA Corporation\\NVIDIA Demos\\%s", szDemoName);

  LONG x;
  HKEY hk; 
  x = RegCreateKeyEx(s_bCurrentUserOnly ? HKEY_CURRENT_USER : HKEY_LOCAL_MACHINE, szKey, 
          0, NULL, REG_OPTION_NON_VOLATILE, KEY_WRITE, NULL, &hk, NULL);
  if ( x!= ERROR_SUCCESS)
    return false;
  
  x = RegSetValueEx(hk, szValueName, NULL, REG_DWORD, (LPBYTE)(&value), sizeof(value));
  return (x == ERROR_SUCCESS);
}

void NVGetStringFromReg(const char* szDemoName, const char* szValueName, char *szBuf, int nBufSize, const char* szDefault)
{
  char szKey[512];
  sprintf(szKey, "SOFTWARE\\NVIDIA Corporation\\NVIDIA Demos\\%s", szDemoName);

  LONG x;
  HKEY hk; 
  x = RegOpenKeyEx(s_bCurrentUserOnly ? HKEY_CURRENT_USER : HKEY_LOCAL_MACHINE, szKey, 
          0, KEY_READ, &hk);
  if (x != ERROR_SUCCESS) {
    strncpy(szBuf, szDefault, nBufSize-1);
    return;
  }

  DWORD type;
  DWORD ret_size = nBufSize-1;
  x = RegQueryValueEx(hk, szValueName, NULL, &type, (LPBYTE)szBuf, &ret_size);
  RegCloseKey(hk);
  if (x != ERROR_SUCCESS || type != REG_SZ)
    strncpy(szBuf, szDefault, nBufSize-1);
}

bool NVWriteStringToReg(const char* szDemoName, const char* szValueName, const char* szStringToWrite)
{
  char szKey[512];
  sprintf(szKey, "SOFTWARE\\NVIDIA Corporation\\NVIDIA Demos\\%s", szDemoName);

  LONG x;
  HKEY hk; 
  x = RegCreateKeyEx(s_bCurrentUserOnly ? HKEY_CURRENT_USER : HKEY_LOCAL_MACHINE, szKey, 
          0, NULL, REG_OPTION_NON_VOLATILE, KEY_WRITE, NULL, &hk, NULL);
  if ( x!= ERROR_SUCCESS)
    return false;
  
  x = RegSetValueEx(hk, szValueName, NULL, REG_SZ, (LPBYTE)szStringToWrite, strlen(szStringToWrite)+1);
  return (x == ERROR_SUCCESS);
}

bool NVGetBinaryFromReg(const char* szDemoName, const char* szValueName, void *dest, unsigned int destBufSize)
{
  char szKey[512];
  sprintf(szKey, "SOFTWARE\\NVIDIA Corporation\\NVIDIA Demos\\%s", szDemoName);

  LONG x;
  HKEY hk; 
  x = RegOpenKeyEx(s_bCurrentUserOnly ? HKEY_CURRENT_USER : HKEY_LOCAL_MACHINE, szKey, 0, KEY_READ, &hk);
  if (x != ERROR_SUCCESS) 
    return false;

  DWORD type;
  DWORD ret_size = destBufSize;
  x = RegQueryValueEx(hk, szValueName, NULL, &type, (LPBYTE)dest, &ret_size);
  RegCloseKey(hk);
  return (x==ERROR_SUCCESS);
}

bool NVWriteBinaryToReg(const char* szDemoName, const char* szValueName, void *data, unsigned int nBytes)
{
  char szKey[512];
  sprintf(szKey, "SOFTWARE\\NVIDIA Corporation\\NVIDIA Demos\\%s", szDemoName);

  LONG x;
  HKEY hk; 
  x = RegCreateKeyEx(s_bCurrentUserOnly ? HKEY_CURRENT_USER : HKEY_LOCAL_MACHINE, szKey, 
          0, NULL, REG_OPTION_NON_VOLATILE, KEY_WRITE, NULL, &hk, NULL);
  if ( x!= ERROR_SUCCESS)
    return false;
  
  x = RegSetValueEx(hk, szValueName, NULL, REG_BINARY, (LPBYTE)data, nBytes);
  return (x == ERROR_SUCCESS);
}

bool NVKeyExists(const char* szDemoName, const char* szValueName)
{
  char szKey[512];
  sprintf(szKey, "SOFTWARE\\NVIDIA Corporation\\NVIDIA Demos\\%s\\%s", szDemoName, szValueName);

  HKEY hk; 
  LONG x = RegOpenKeyEx(s_bCurrentUserOnly ? HKEY_CURRENT_USER : HKEY_LOCAL_MACHINE, szKey, 0, KEY_READ, &hk);
  RegCloseKey(hk);
  return (x == ERROR_SUCCESS);
}

bool NVDeleteKey(const char* szDemoName, const char* szValueName)
{
  char szKey[512];
  sprintf(szKey, "SOFTWARE\\NVIDIA Corporation\\NVIDIA Demos\\%s", szDemoName);

  LONG x;
  HKEY hk; 
  x = RegOpenKeyEx(s_bCurrentUserOnly ? HKEY_CURRENT_USER : HKEY_LOCAL_MACHINE, szKey, 0, KEY_SET_VALUE, &hk);
  if (x != ERROR_SUCCESS) 
    return false;

  x = RegDeleteValue(hk, szValueName);
  RegCloseKey(hk);
  return (x != ERROR_SUCCESS);
}

void SetScrollLock(bool bNewState)
{
	if (bNewState != (bool)(GetKeyState(VK_SCROLL) & 1))
	{
		// Simulate a key press
		keybd_event( VK_SCROLL,
					  0x45,
					  KEYEVENTF_EXTENDEDKEY | 0,
					  0 );

		// Simulate a key release
		keybd_event( VK_SCROLL,
					  0x45,
					  KEYEVENTF_EXTENDEDKEY | KEYEVENTF_KEYUP,
					  0);
	}
}


//------------[ FPS MONITORING FUNCTIONS ]------------

CFpsMonitor::CFpsMonitor(int nFramesToTrack) {
  assert(nFramesToTrack >= 256);
  assert(nFramesToTrack <= 16384);

  m_nLogSize = nFramesToTrack;
  m_fFpsLog = new float[nFramesToTrack];
  Reset();
}

CFpsMonitor::~CFpsMonitor() {
  delete m_fFpsLog;
  m_fFpsLog = NULL;
}

void CFpsMonitor::Reset() {
  m_nFramesSinceReset = -3; // don't count the first few frames
  m_fFpsLogSum = 0;
  m_fDampedFps = 30.0f;
}

float CFpsMonitor::Tick(float unpausable_delta_t) {
  
  if (unpausable_delta_t < 0.001f)
    return m_fDampedFps;
  if (m_nFramesSinceReset < 0) {
    m_nFramesSinceReset++;
    return m_fDampedFps;
  }

  float fps = 1.0f / unpausable_delta_t;
  
  // if demo stalled for > 10X the length of one frame
  // [like from a break in the debugger!], filter that out:
  if (m_nFramesSinceReset > 16 && fps < m_fDampedFps*0.1f)
    return m_fDampedFps;

  int i = m_nFramesSinceReset % m_nLogSize;
  if (m_nFramesSinceReset >= m_nLogSize)
    m_fFpsLogSum -= m_fFpsLog[i];
  m_fFpsLogSum += fps;
  m_fFpsLog[i] = fps;
  m_fDampedFps = m_fFpsLogSum / min(m_nFramesSinceReset+1, m_nLogSize);
  m_nFramesSinceReset++;
  return m_fDampedFps;
}




int GetIntFromDistribution(int* distribution, int nIndices)
{
  // say you pass in a distribution of { 10, 15, 25 }.
  // this will pass back either 0 (20% chance), 1 (30% chance), or 2 (50% chance).

  int sum = 0;
  for (int i=0; i<nIndices; i++)
    sum += distribution[i];
  int x = rand() % sum;
  int sumSoFar = 0;
  for (int i=0; i<nIndices; i++) {
    sumSoFar += distribution[i];
    if (x < sumSoFar)
      return i;
  }
  assert(0);
  return 0;
}

int RandFromProb(float* cum_prob_array, int array_size)
{
  // given an input array of floats describing a *cumulative* distibution
  //     such as { 0.2, 0.31, 0.77, 0.83, 1.0 }
  // returns an int of 0..4 (in above case) respecting the distribution.
  float f = FRAND;
  for (int i=array_size-2; i>=0; i--)
    if (f > cum_prob_array[i])
      return i+1;
  return 0;
}





